﻿const net = require('net');
const PORT = 8765;
const FID_STOP = 0;

function Client (cb) {
  this.socket = net.connect({port: PORT}, setImmediate.call(null, cb));
  this.buf = Buffer(10*1024*1024);
  this.bufpos = 0;
  this.cbs = [];
  
  this.socket.on('data', chunk => {
    chunk.forEach((b, i) => {
      const cb = this.cbs.shift();
      if (!cb) throw new Error('unexpected packet ' + b + ' ' + i + '/' + chunk.length);
      cb(b);
    });
  });
}

Client.prototype.messages = function (msgs, one, cb) {
  console.log('messages ' + (one ? 'one chunk' : 'annoying'));
  msgs.forEach((msg, i) => {
    console.log('message: ' + JSON.stringify(msg));
    if (typeof msg === 'number') {
      this.buf.writeUInt8(msg & 0xff, this.bufpos++);
      this.cbs.push(b => {
        console.log('response to ' + msg, b);
        if (i == msgs.length - 1) {
          if (this.bufpos) throw new Error('received before finishing sending');
          console.log('messages: sent and received');
          cb();
        }
      });
      return;
    }
    
    this.buf.writeUInt8(1, this.bufpos++);
    msg.sort(() => Math.random() - 0.5).forEach(entry => {
      const fid = entry[0], v = Buffer(entry[1]);
      this.buf.writeUInt8(fid,this.bufpos++);
      this.buf.writeUInt32BE(v.length, this.bufpos);
      this.bufpos += 4;
      v.copy(this.buf, this.bufpos);
      this.bufpos += v.length;
    });
    this.buf.writeUInt8(FID_STOP, this.bufpos++);
    this.cbs.push(b => {
      console.log('response to ' + JSON.stringify(msg), b);
      if (i == msgs.length - 1) {
        if (this.bufpos) throw new Error('received before finishing sending');
        console.log('messages: sent and received');
        cb();
      }
    });
  });
  
  if (one) {
    console.log('send in one chunk: ' + this.bufpos + ' bytes');
    this.socket.write(this.buf.slice(0, this.bufpos));
    console.log('sent!');
    this.bufpos = 0;
    return;
  }
  
  var sent = 0;
  const sendNext = () => {
    var n = Math.random() * (this.bufpos - sent + 1) | 0;
    if (n > 0) {
      console.log('send ' + n + ' bytes (' + sent + '..' + (sent+n) + '/' + this.bufpos + ')');
      this.socket.write(this.buf.slice(sent, sent + n));
      sent += n;
    }
    
    if (sent >= this.bufpos) {
      console.log('sent!');
      this.bufpos = 0;
      return;
    }
    
    const to = Math.random() * 2000 | 0;
    console.log('sleeping for ' + to + 'ms');
    setTimeout(sendNext, to);
  };
  
  sendNext();
};

Client.prototype.end = function () {
  this.socket.end();
};

var msgs = [
  0,
  2,
  [[1, 'hello'], [2, 'world'], [3 , 'ololo']],
  [[5, 'hallo'], [7, 'world'], [14, 'oLOLo']],
  [[6, 'hQllo'], [6, 'world'], [6 , 'oloso']],
  [[7, 'hlo'],   [7, 'woRLD'], [7 , 'olsslo']],
];

var clnt = new Client (() => {
  console.log('------------ test 1 -------------');
  clnt.messages(msgs, true, () => {
    clnt.end();
    
    clnt = new Client (() => {
      console.log('------------ test 2 -------------');
      clnt.messages(msgs, false, () => {
        clnt.end();
        console.log('------------ end -------------');
      });
    });
  })
});
