﻿const net = require('net');
const PORT = 8765;
const FID_STOP = 0;

var lastcid = 0;
const srv = net.createServer(client => {
  const cid = ++lastcid;
  var goal = null;
  var received = [];
  var cbs = [], goals = [];
  
  const finish = error => {
    var cbs1 = cbs;
    cbs = null;
    goals = null;
    cbs1.forEach(function (cb) {
      cb(error);
    });
  };
  
  const onByte = b => {
    received.push(b);
    
    if (goal == null) {
      goal = goals.shift();
      if (goal == null) return;
    }
   
    if (received.length >= goal) {
      goal = goals.shift();
      const res = received;
      received = [];
      const cb = cbs.shift();
      cb(null, res);
    }
  };
  
  const read = (n, cb) => {
    if (goal == null && received.length >= n) {
      const res = received.slice(0, n);
      received = received.slice(n);
      cb(null, res);
    } else {
      goals.push(n);
      cbs.push(cb);
    }
  };
  
  console.log('client ' + cid + ' connected');
  client.on('end', () => {
    console.log('client ' + cid + ' disconnected');
    finish(new Error('disconnected'));
  }).on('data', chunk => {
    console.log('client ' + cid + ' data ('+chunk.length+')', chunk);
    chunk.forEach(onByte);
  }).on('error', err => {
    console.log('client ' + cid + ' error', err);
    finish(err);
  });
  
  (function readMessage () {
    read(1, function (err, cmd_) {
      if (err) return void console.log('client ' + cid + ' read error: ' + err.message);
      const cmd = cmd_[0];
      console.log('client ' + cid + ' parsed cmd=' + cmd);

      switch (cmd) {
        case 0: // hello
          client.write(Buffer([0]));
          console.log('client ' + cid + ' success');
          readMessage();
          break;
        
        case 1: // msg
          var values = Object.create(null);
          (function readFields () {
            read(1, function (err, fid_) {
              if (err) return void console.log('client ' + cid + ' read error: ' + err.message);
              const fid = fid_[0];
              
              if (fid === FID_STOP) {
                console.log('client ' + cid + ' parsed stop');
                console.log('client ' + cid + ' MESSAGE', JSON.stringify(values, null, 2));
                client.write(Buffer([0]));
                console.log('client ' + cid + ' success');
                readMessage();
                return;
              }
              
              read(4, function (err, fsize_) {
                if (err) return void console.log('client ' + cid + ' read error: ' + err.message);
                const fsize = Buffer(fsize_).readUInt32BE(0);
                
                read(fsize, function (err, fdata) {
                  if (err) return void console.log('client ' + cid + ' read error: ' + err.message);
                  const fvalue = String(Buffer(fdata));
                  console.log('client ' + cid + ' parsed fid=' + fid + ' len=' + fsize + ' val=' + fvalue);
                  if (!(fid in values)) values[fid] = [];
                  values[fid].push(fvalue);
                  readFields();
                });
              });
            });
          })();
          break;
        
        case 2: // updateConfig
          client.write(Buffer([0]));
          console.log('client ' + cid + ' success');
          readMessage();
          break;
        
        default: // error
          client.end();
          console.log('client ' + cid + ' GTFO');
      }
    });
  })();
}).listen(PORT, () => {
  console.log('server: listening');
});